#!/usr/bin/env python
# -*- coding: utf-8 -*-
### BEGIN LICENSE
# Large parts of this code are based on otrverwaltung++ by Markus Liebl.
# Copyright (C) 2013 Markus Liebl <lieblm@web.de>

# Modified and extended by Guenter Kreidl for use on Raspberry Pi
# Copyright of modifications and extensions 2018 by Günter Kreidl <gkreidl@krefelder-life.de>
# Version 0.5
#
#This program is free software: you can redistribute it and/or modify it
#under the terms of the GNU General Public License version 3, as published
#by the Free Software Foundation.
#
#This program is distributed in the hope that it will be useful, but
#WITHOUT ANY WARRANTY; without even the implied warranties of
#MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
#PURPOSE.  See the GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License along
#with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE

import re
import os
import sys
import subprocess
import bisect
import shutil
import ConfigParser
import time

def __error(message_text):
    print message_text

def handle_error(error_cb, message):
    if error_cb:
        error_cb(message)
    else:
        print message
                    
def remove_file(filename, error_cb=None):
    """ Entfernt die angegebene Datei. """

#    print "[Fileoperations] Remove ", filename
    try:
        os.remove(filename)
    except Exception, e:
        handle_error(error_cb, "Fehler beim Löschen von %s (%s)." % (filename, e))
    
def rename_file(old_filename, new_filename, error_cb=None):
    """ Benennt eine Datei um. Wenn die Datei bereits existiert, wird der neue Name um eine Zahl erweitert. """
       
    if old_filename == new_filename:
        handle_error(error_cb, "Umbenennen: Die beiden Dateinamen stimmen überein! (%s)" % old_filename)
        return

    new_filename = make_unique_filename(new_filename)
    
#    print "[Fileoperations] Rename %s to %s" % (old_filename, new_filename)
        
    try:        
        os.rename(old_filename, new_filename)
    except Exception, e:
        handle_error(error_cb, "Fehler beim Umbenennen von %s nach %s (%s)." % (old_filename, new_filename, e))
        return
    
    return new_filename

def make_unique_filename(filename):
    """ Gleicht den gegebenen Dateinamen an, sodass  """
    new_filename = filename
    count = 1
    while os.path.exists(new_filename):
        path, extension = os.path.splitext(filename)
        new_filename = "%s.%i%s" % (path, count, extension)
        count += 1

    return new_filename

class Cut_action:
    ASK             = 0
    BEST_CUTLIST    = 1
    CHOOSE_CUTLIST  = 2
    MANUALLY        = 3
    LOCAL_CUTLIST   = 4
    
class Format:
    AVI = 0
    HQ  = 1
    MP4 = 2
    HD  = 3
    AC3 = 4
    
class Program:
    AVIDEMUX    = 0
    VIRTUALDUB  = 1
    CUT_INTERFACE = 2
    SMART_MKVMERGE = 3
    
class Config:
    def __init__(self):        
        self.__fields = {
            "general": {
                "choose_cutlists_by": 1,
                "cut_action": 0,
                "cut_hqs_by": "SmartMKVmerge",
                "cutlist_hash": "eb34eeb9e90006ae43525e366ad0ae96",
                "cutlist_username": "",
                "delete_cutlists": False,
                "delete_otrkeys": False,
                "preview_delay": 3,
                "preview_cuts": False,
                "email": "",
                "password": "",
                "folder_cut_avis": "/home/pi/otr/cut",
                "folder_new_otrkeys": "/home/pi/Downloads",
                "folder_trash_avis": "/home/pi/otr/trash",
                "folder_uncut_avis": "/home/pi/otr/decoded",
                "folder_converted_avis": "/home/pi/otr/converted",
                "merge_ac3s": True,
                "merge_ac3s_by": "mkvmerge",
                "play_cut_videos": False,
                "rename_cut": False,
                "server": "http://cutlist.at/",
                "x264vfw_hd_string": "--tune film --direct auto --force-cfr --rc-lookahead 60 --b-adapt 2 --weightp 0",
                "x264vfw_hq_string": "--tune film --direct auto --force-cfr --rc-lookahead 60 --b-adapt 2 --aq-mode 2 --weightp 0",
                "x264vfw_mp4_string": "--force-cfr --profile baseline --preset medium --trellis 0",
                "videoplayers": ["/usr/local/bin/omxplayergui","/usr/bin/vlc"]
            },
            "programs": {
                "decoder": "otrpidecoder",
                "ffmpeg": "ffmpeg",
                "ffmsindex": "ffmsindex",
                "ffprobe": "ffprobe",
                "mediainfo": "mediainfo",
                "mkvmerge": "mkvmerge",
                "sox": "sox",
                "wget": "wget",
                "x264": "x264"
            },
            "smartmkvmerge": {
                "encoder_engine": "x264",
                "ffmpeg_avi_mpeg4_options": "-c:v mpeg4 -vtag DX50 -qscale:v 5 -g 300 -f avi",
                "ffmpeg_hd_x264_options": "crf=23:deblock=-1/-1:psy-rd=1/0.15:direct=auto:force-cfr=1:b-adapt=2:rc-lookahead=60:weightp=0:aq-mode=2",
                "ffmpeg_hq_x264_options": "crf=23:deblock=-1/-1:psy-rd=1/0.15:direct=auto:force-cfr=1:b-adapt=2:rc-lookahead=60:weightp=0",
                "ffmpeg_mp4_x264_options": "crf=23:force-cfr=1:trellis=0",
                "first_audio_stream": "originale MP3 Spur kopieren",
                "normalize_audio": False,
                "second_audio_stream": "originale AC3 Spur kopieren",
                "remux_to_mp4": False,
                "single_threaded": True,
                "single_threaded_automatic": False,
                "workingdir": "/home/pi/otr/tmp",
                "x264_hd_string": "--tune film --direct auto --force-cfr --rc-lookahead 60 --b-adapt 2 --weightp 0",
                "x264_hq_string": "--tune film --direct auto --force-cfr --rc-lookahead 60 --b-adapt 2 --aq-mode 2 --weightp 0",
                "x264_mp4_string": "--force-cfr --profile baseline --preset medium --trellis 0"
            }
        }
    def get(self, category, option):
        value = self.__fields[category][option]
        return value

    def setvalue(self,category,option,value):
        self.__fields[category][option] = value

    def get_program(self, program):
        value = self.__fields['programs'][program]
        return value

    def load(self,filepath):
        if os.path.exists(filepath):
            f = file(filepath,'rb')
            cf = f.read().decode('utf8')
            f.close()
            try:
                self.__fields = eval(cf)
            except:
                print "Config read error"
                pass
    def save(self,filepath):
        cf = str(self.__fields)  #.replace('},','},\n').replace("',","',\n")
        f = file(filepath,'wb')
        cf = f.write(cf.encode('utf8'))
        f.close()

class Cutlist:
    
    def __init__(self,cutlistfile):
    
        # cutlist.at xml-output
        self.id = 0
        self.author = ''
        self.ratingbyauthor = 0
        self.rating = 0
        self.ratingcount = 0
        self.countcuts = 0        # !!! 
        self.actualcontent = ''
        self.usercomment = ''
        self.filename = ''
        self.withframes = 0
        self.withtime = 0
        self.duration = ''
        self.errors = ''
        self.othererrordescription = ''
        self.downloadcount = 0
        self.autoname = ''
        self.filename_original = ''

        # additions in cutlist file
        self.wrong_content = 0
        self.missing_beginning = 0
        self.missing_ending = 0
        self.other_error = 0
        self.suggested_filename = ''
        self.intended_app = ''
        self.intended_version = ''
        self.smart = 0
        self.fps = 0

        # own additions
        self.app = 'rpiotrtool'
        self.errors = False
        self.cuts_seconds = [] # (start, duration) list
        self.cuts_frames = [] # (start, duration) list
        self.local_filename = cutlistfile
    
    def _check_string(self, string):
        try:
            res = string.decode('latin1').encode('utf-8')
        except:
            print "No latin1 string, trying utf-8"
            try:
                res = string.decode('utf-8').encode('utf-8')
            except:
                print "Malformed string"
                res = ""
        
        return res
    
    def read_from_file(self):
        config_parser = ConfigParser.SafeConfigParser()
        
        try:
            config_parser.read(self.local_filename)
            
            self.filename = self._check_string(config_parser.get('Info', 'SuggestedMovieName'))
            self.author = self._check_string(config_parser.get('Info', 'Author'))
            self.ratingbyauthor = int(config_parser.get('Info', 'RatingByAuthor'))
            self.rating = 0
            self.ratingcount = 0
            self.usercomment = self._check_string(config_parser.get('Info', 'UserComment'))
            self.countcuts = int(config_parser.get('General', 'NoOfCuts'))
            self.actualcontent = self._check_string(config_parser.get('Info', 'ActualContent'))
            self.filename_original = self._check_string(config_parser.get('General', 'ApplyToFile'))
        except:
            print "Malformed cutlist: ", self.local_filename
      
    def read_cuts(self):
        """ Reads cuts from local_filename.
            
            Returns: error message, otherwise None """
                            
        config_parser = ConfigParser.SafeConfigParser()        
         
        try:            
            config_parser.read(self.local_filename)
        except ConfigParser.ParsingError, message:
            print "Malformed cutlist: ", message

        try:
            noofcuts = int(config_parser.get("General", "NoOfCuts"))
            
            for count in range(noofcuts):
                cut = "Cut" + str(count)
            
                if config_parser.has_option(cut, "StartFrame") and config_parser.has_option(cut, "DurationFrames"):
                    start_frame = int(config_parser.get(cut, "StartFrame"))
                    duration_frames = int(config_parser.get(cut, "DurationFrames"))
                    if duration_frames > 0:
                        self.cuts_frames.append((start_frame, duration_frames))                    

                start_second = float(config_parser.get(cut, "Start"))
                duration_seconds = float(config_parser.get(cut, "Duration"))                                    
                if duration_seconds > 0:
                    self.cuts_seconds.append((start_second, duration_seconds))     
        except ConfigParser.NoSectionError, message:
            return "Fehler in Cutlist: " + str(message)
        except ConfigParser.NoOptionError, message:
            return "Fehler in Cutlist: " + str(message)

class BaseAction:    
    """ Base class for actions."""    
     
    def __init__(self):
        self.update_list = True
    
    def do(self):
        raise Exception("Override this method!")

class Cut(BaseAction):

    def __init__(self, config):
        self.update_list = True
        self.config = config
        
    def cut_file_by_cutlist(self):
        raise Exception("Override this method!")
    
    def create_cutlist(self):
        raise Exception("Override this method!")
        
    def get_format(self, filename):
        root, extension = os.path.splitext(filename)
    
        if extension == '.avi':
            bframe_delay = 2
            if os.path.splitext(root)[1] == '.HQ':
                format = Format.HQ
                ac3name = os.path.splitext(root)[0] + ".HD.ac3"
            elif os.path.splitext(root)[1] == '.HD':
                format = Format.HD
                ac3name = root + ".ac3"
            else:
                format = Format.AVI
                ac3name = root + ".HD.ac3"
        elif extension == '.mp4':
            bframe_delay = 0
            if os.path.splitext(root)[1] == '.HQ':
                format = Format.HQ
                ac3name = os.path.splitext(root)[0] + ".HD.ac3"
            elif os.path.splitext(root)[1] == '.HD':
                format = Format.HD
                ac3name = root + ".ac3"
            else:
                format = Format.MP4
                ac3name = root + ".HD.ac3"
        elif extension == '.mkv':
            bframe_delay = 2
            if os.path.splitext(root)[1] == '.HQ':
                format = Format.HQ
                ac3name = os.path.splitext(root)[0] + ".HD.ac3"
            elif os.path.splitext(root)[1] == '.HD':
                format = Format.HD
                ac3name = root + ".ac3"
            else:
                format = Format.AVI
                ac3name = root + ".HD.ac3"
        elif extension == '.ac3':
                format = Format.AC3
                ac3name = root
        else:
            return -1, None
    
        if os.path.isfile(ac3name):
            return format, ac3name, bframe_delay
        else:
            return format, None, bframe_delay

    def get_program(self, filename, manually=False):
        if manually:
            programs = { Format.AVI : self.config.get('general', 'cut_avis_man_by'),
                         Format.HQ  : self.config.get('general', 'cut_hqs_man_by'),
                         Format.HD  : self.config.get('general', 'cut_hqs_man_by'),
                         Format.MP4 : self.config.get('general', 'cut_mp4s_man_by') }
        else:
            programs = { Format.AVI : self.config.get('general', 'cut_avis_by'),
                         Format.HQ  : self.config.get('general', 'cut_hqs_by'),
                         Format.HD  : self.config.get('general', 'cut_hqs_by'),
                         Format.MP4 : self.config.get('general', 'cut_mp4s_by') }
                     
        format, ac3, bframe_delay = self.get_format(filename)

        if format < 0:
            return -1, "Format konnte nicht bestimmt werden/wird noch nicht unterstützt.", False

        if format == Format.AC3:
            return -1, "AC3 wird automatisch mit der HD.avi verarbeitet und nicht einzeln geschnitten.", False

        config_value = programs[format]

        if 'SmartMKVmerge' in config_value:
            return Program.SMART_MKVMERGE, config_value, ac3
        else:
            return -2, "Programm '%s' konnte nicht bestimmt werden. Es werden VirtualDub und Avidemux unterstützt." % config_value, False

    def generate_filename(self, filename, forceavi=0):
        """ generate filename for a cut video file. """

        root, extension = os.path.splitext(os.path.basename(filename))
        if forceavi == 1:
            extension = '.avi'
        new_name = root + "-cut" + extension

        cut_video = os.path.join(self.config.get('general', 'folder_cut_avis'), new_name)

        return cut_video

    def mux_ac3(self, filename, cut_video, ac3_file, cutlist):	# cuts the ac3 and muxes it with the avi into an mkv
        mkvmerge = self.config.get_program('mkvmerge')
        root, extension = os.path.splitext(filename)
        mkv_file = os.path.splitext(cut_video)[0] + ".mkv"
        # env
        my_env = os.environ.copy()
        my_env["LANG"] = "C"

        # creates the timecodes string for splitting the .ac3 with mkvmerge
        timecodes = (','.join([self.get_timecode(start) + ',' + self.get_timecode(start+duration) for start, duration in cutlist.cuts_seconds]))
        # splitting .ac3. Every second fragment will be used.
        try:
            blocking_process = subprocess.Popen([mkvmerge, '--ui-language',  'en_US',  "--split", "timecodes:" + timecodes, "-o", root + "-%03d.mka", ac3_file ], stdout=subprocess.PIPE, stderr=subprocess.STDOUT, universal_newlines=True,  env=my_env)
        except OSError as e:
            return None,  e.strerror + ": " + mkvmerge
        return_value = blocking_process.wait()
        # return_value=0 is OK, return_value=1 means a warning. Most probably non-ac3-data that has been omitted.
        if return_value != 0 and return_value != 1:
            return None, None, str(return_value)

        if len(cutlist.cuts_seconds) == 1:              # Only the second fragment is needed. Delete the rest.
            rename_file(root + "-002.mka", root + ".mka")
            remove_file(root + "-001.mka")
            if os.path.isfile(root + "-003.mka"):
                remove_file(root + "-003.mka")

        else:                                           # Concatenating every second fragment.
            command = [mkvmerge, "-o", root + ".mka", root + "-002.mka"]
            command[len(command):] = ["+" + root + "-%03d.mka" % (2*n) for n in range(2,len(cutlist.cuts_seconds)+1)]
            try:
                blocking_process = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, universal_newlines=True,  env=my_env)
            except OSError as e:
                return None,  e.strerror + ": " + mkvmerge
            return_value = blocking_process.wait()
            if return_value != 0:               # There should be no warnings here
                return None, None, str(return_value)

            for n in range(1,2*len(cutlist.cuts_seconds)+2):    # Delete all temporary audio fragments
                if os.path.isfile(root + "-%03d.mka" % n):
                    remove_file(root + "-%03d.mka" % n)

        # Mux the cut .avi with the resulting audio-file into mkv_file
        # TODO: Is there some way to pass possible warnings to the conclusion dialog?
        try:
            blocking_process = subprocess.Popen([mkvmerge, "-o", mkv_file, cut_video, root + ".mka"], stdout=subprocess.PIPE, stderr=subprocess.STDOUT, universal_newlines=True,  env=my_env)
        except OSError as e:
            return None,  e.strerror + ": " + mkvmerge
        return_value = blocking_process.wait()
        if return_value != 0 and return_value != 1:
            return None, None, str(return_value)

        remove_file(root + ".mka")       # Delete remaining temporary files
        remove_file(cut_video)
        return mkv_file, ac3_file, None

    def get_timecode(self, time):           # converts the seconds into a timecode-format that mkvmerge understands
        minute, second = divmod(int(time),60)		# discards milliseconds
        hour, minute = divmod(minute, 60)
        second = time - minute * 60 - hour * 3600	# for the milliseconds
        return "%02i:%02i:%f" % (hour, minute, second)

    def analyse_mediafile(self, filename):
        """ Gets fps, dar, sar, number of frames and id of the ac3_stream of a movie using ffmpeg.
            Returns without error:
                fps, dar, sar, max_frames, ac3_stream, None
            with error:
                None, None, None, None, None, error_message """
                
        try:
            process = subprocess.Popen([self.config.get_program('ffmpeg'), "-i", filename], stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        except OSError:
            return None, None, None, None, None,"FFMPEG (static) konnte nicht ausgeführt werden!"
    
        log = process.communicate()[0]
    
        video_infos_match = re.compile(r".*(Duration).*(\d{1,}):(\d{1,}):(\d{1,}.\d{1,}).*|.*(SAR) (\d{1,}:\d{1,}) DAR (\d{1,}:\d{1,}).*\, (\d{2,}\.{0,}\d{0,}) tbr.*|.*(Stream).*(\d{1,}:\d{1,}).*Audio.*ac3.*")
        seconds = 0
        ac3_stream = fps = dar = sar = None
    
        for line in log.split('\n'):
            m = re.search(video_infos_match,line)
    
            if m:
                if "Duration" == m.group(1):
                    try:
                        seconds = float(m.group(2))*3600 + float(m.group(3))*60 + float(m.group(4))
                    except ValueError:
                        return None, None, None, None,  "Dauer des Film konnte nicht ausgelesen werden."
                elif "SAR" == m.group(5):
                    try:
                        sar = m.group(6)
                        dar = m.group(7)
                        fps = float(m.group(8))
                    except ValueError:
                        return None, None, None, None,  "Video Stream Informationen konnte nicht ausgelesen werden."                        
                elif "Stream" == m.group(9):
                    ac3_stream = m.group(10)
            else:
                pass
                
        if seconds != 0 and fps != None and sar != None and dar != None:
            max_frames = seconds * fps
            return fps,  dar,  sar,  max_frames, ac3_stream,  None
    
        return None, None, None, None, None, "Es konnten keine Video Infos der zu bearbeitenden Datei ausgelesen werden."
    
    def get_keyframes_from_file(self, filename):
        """ returns keyframe list - in frame numbers"""
        
        if not os.path.isfile(filename + '.ffindex_track00.kf.txt'):
            try:
                command = [self.config.get_program('ffmsindex'),  '-p', '-f', '-k',  filename ]
                ffmsindex = subprocess.call(command)
            except OSError:
                return None, "ffmsindex konnte nicht aufgerufen werden."
    
        if os.path.isfile(filename + '.ffindex_track00.kf.txt'):
            filename_keyframes = filename + '.ffindex_track00.kf.txt'
        elif os.path.isfile(filename + '.ffindex_track01.kf.txt'):
            filename_keyframes = filename + '.ffindex_track01.kf.txt'
        elif os.path.isfile(filename + '.ffindex_track02.kf.txt'):
            filename_keyframes = filename + '.ffindex_track02.kf.txt'
        else:
            filename_keyframes = None

        try:
            index = open(filename_keyframes, 'r')
        except IOError:
            return None,  "Keyframe File von ffmsindex konnte nicht geöffnet werden."
        index.readline()
        index.readline()
        try:
            list =[int(i) for i in index.read().splitlines()]
        except ValueError:
            return None,  "Keyframes konnten nicht ermittelt werden."
        index.close()
        if os.path.isfile(filename + '.ffindex'):
            remove_file(filename +'.ffindex')
            
        return list,  None
    
    def get_keyframe_in_front_of_frame(self, keyframes,  frame):
        """Find keyframe less-than to frame."""
    
        i = bisect.bisect_left(keyframes, frame)
        if i:
            return keyframes[i-1]
        raise ValueError
    
    def get_keyframe_after_frame(self,  keyframes, frame):
        """Find keyframe greater-than to frame."""
       
        i = bisect.bisect_right(keyframes, frame)
        if i != len(keyframes):
            return keyframes[i]
        raise ValueError 
            
    def complete_x264_opts(self, x264_opts, filename):
        """Analyse filename and complete the x264 options 
        returns 
          x264_opts  x264 options
          x264_core  x264 core version
        """
        
        bt709 = ['--videoformat', 'pal', '--colorprim', 'bt709', '--transfer', 'bt709', '--colormatrix',  'bt709']
        bt470bg = ['--videoformat', 'pal', '--colorprim', 'bt470bg', '--transfer', 'bt470bg', '--colormatrix',  'bt470bg']
        
        try:
            blocking_process = subprocess.Popen([self.config.get_program('mediainfo'), filename], stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True)
        except OSError as e:
            return None, "Fehler: %s Filename: %s Error: %s" % str(e.errno),  str(e.filename),  str(e.strerror)
        except ValueError as e:
            return None,  "Falscher Wert: %s" % str(e)
        
        while True:
            line = blocking_process.stdout.readline()

            if line != '':
                if 'x264 core' in line:
                    #print line
                    try:
                        x264_core = int(line.strip().split(' ')[30])
                        #print x264_core
                    except ValueError as e:
                        continue
                    except IndexError as e:
                        continue
                elif 'Color primaries' in line and '709' in line:
                    x264_opts.extend(bt709)
                elif 'Color primaries' in line and '470' in line:
                    x264_opts.extend(bt470bg)
                elif 'Format profile' in line and '@L' in line:
                    try:
                        level = ['--level', str(float(line.strip().split('L')[1]))] # test for float
                        profile = ['--profile', line.strip().split('@L')[0].split(':')[1].lower().lstrip()]
                    except ValueError as e:
                        continue
                    except IndexError as e:
                        continue
                    x264_opts.extend(profile)
                    x264_opts.extend(level)
                elif 'Frame rate' in line:
                    try:
                        fps = ['--fps', str(float(line.strip().split(' ')[3]))]
                    except ValueError as e:
                        continue
                    except IndexError as e:
                        continue
                    x264_opts.extend(fps)
            else:
                break
        return x264_opts,  x264_core

    def show_progress(self, blocking_process):
        progress_match = re.compile(r".*(?<=\[|\ )(\d{1,}).*%.*")
        time_match = re.compile(r".*(\d{1,}):(\d{1,}):(\d{1,}.\d{1,}).*")
        mp4box_match = re.compile(r".*\((\d{2,})\/\d{2,}\).*")
        
        max_sec = 0.0
                
        while True:
            try:
                #line = blocking_process.stdout.readline()
                line = blocking_process.stdout.read()
            except:
                break
            if line == '':
                break
            else:
                time.sleep(0.5)
                continue

    def get_norm_volume(self, filename, stream):
        """ Gets the volume correction of a movie using ffprobe. 
            Returns without error:              
                        norm_vol, None
                    with error:
                        1.0, error_message """
            
        try:
            process1 = subprocess.Popen([path.get_tools_path('intern-ffprobe'), '-v',  'error','-of','compact=p=0:nk=1','-drc_scale','1.0','-show_entries','frame_tags=lavfi.r128.I','-f','lavfi','amovie='+filename+':si='+stream+',ebur128=metadata=1'], stdout=subprocess.PIPE)       
        except OSError:
            return "1.0", "FFMPEG wurde nicht gefunden!"            
                
        log = process1.communicate()[0]
            
    
        loudness = ref = -23
        for line in log.splitlines():
            sline = line.rstrip()
            if sline:
                loudness = sline
                adjust = ref - float(loudness)
        if adjust:
            return str(adjust)+'dB',  None
        else:
            return "1.0", "Volume konnte nicht bestimmt werden."

    def available_cpu_count(self):
        """ Number of available virtual or physical CPUs on this system, i.e.
        user/real as output by time(1) when called with an optimally scaling
        userspace-only program"""
    
        # cpuset
        # cpuset may restrict the number of *available* processors
        try:
            m = re.search(r'(?m)^Cpus_allowed:\s*(.*)$',
                          open('/proc/self/status').read())
            if m:
                res = bin(int(m.group(1).replace(',', ''), 16)).count('1')
                if res > 0:
                    return res
        except IOError:
            pass
    
        # Python 2.6+
        try:
            import multiprocessing
            return multiprocessing.cpu_count()
        except (ImportError, NotImplementedError):
            pass
    
        # Linux
        try:
            res = open('/proc/cpuinfo').read().count('processor\t:')
    
            if res > 0:
                return res
        except IOError:
            pass
    
        # Other UNIXes (heuristic)
        try:
            try:
                dmesg = open('/var/run/dmesg.boot').read()
            except IOError:
                dmesgProcess = subprocess.Popen(['dmesg'], stdout=subprocess.PIPE)
                dmesg = dmesgProcess.communicate()[0]
    
            res = 0
            while '\ncpu' + str(res) + ':' in dmesg:
                res += 1
    
            if res > 0:
                return res
        except OSError:
            pass
    
        raise Exception('Can not determine number of CPUs on this system')
        
    def meminfo(self):
        """ return meminfo dict """
        
        meminfo=dict()
        with os.popen('cat /proc/meminfo') as f:
            for l in f:
                x = l.split()
                meminfo[x[0][:-1]] = int(x[1])
        return meminfo

class CutSmartMkvmerge(Cut):

    def __init__(self, config):
        self.update_list = True
        self.config = config
        self.encode_nr = 0                                     # index for encoded video parts - for the smart rendering simulation
        self.copy_nr = 0                                         # index for copied video parts - for the smart rendering simulation
        self.workingdir = '/tmp'
        self.video_files = []                                     # temporary video files
        self.audio_files = []                                    # temporary audio files
        self.rawstreams = {}                                # temporary eac3to files
        
    def __del__(self):
        # clean up
        try:
            if os.path.isfile (self.workingdir + '/audio_copy.mkv'):
                os.remove(self.workingdir + '/audio_copy.mkv')
            if os.path.isfile (self.workingdir + '/x264.index'):
                os.remove(self.workingdir + '/x264.index')
            if os.path.isfile (self.workingdir + '/video_copy.mkv'):
                os.rename(self.workingdir + '/video_copy.mkv', self.workingdir + '/video_copy-001.mkv')
            for n in self.video_files + self.audio_files:
                if os.path.isfile(n.lstrip('+')):
                    os.remove(n.lstrip('+'))
            if os.path.isfile (self.workingdir + '/audio_encode.mkv'):
                os.remove(self.workingdir + '/audio_encode.mkv')
            for index in sorted(self.rawstreams.keys()):
                if os.path.isfile(self.workingdir + '/' + self.rawstreams[index]):
                    os.remove(self.workingdir + '/' + self.rawstreams[index])
        except:
            pass
        

    def cut_file_by_cutlist(self, filename, cutlist=None, program_config_value=None):
        """ Cuts a otr file with x264 and mkvmerge frame accurate. 
            returns: name of cut video, error_message """
        # configuration
        videolist = []                                              # result list for smart rendering simulation
        audio_import_files = [filename]                 # otr files which have audio streams  and needs to be cutted (e.g. OTR avi and ac3) 
        process_list = []                                        # list of started processes
        mkvmerge_list = []                                     # list of started mkvmerge processes
        video_splitframes = ''                               # mkvmerge split string for cutting the video at keyframes 
        audio_timecodes = ''                                # mkvmerge split timecodes for cutting the audio
        ac3_file = None                                         # AC3 source file
        warning_msg = None
        mkvmerge =  self.config.get_program('mkvmerge')
        x264 = self.config.get_program('x264')
        ffmpeg = self.config.get_program('ffmpeg')
        encoder_engine = self.config.get('smartmkvmerge', 'encoder_engine')
        # env
        my_env = os.environ.copy()
        my_env["LANG"] = "C"
        my_env["LC_COLLATE"] = "C"
        my_env["LC_ALL"] = "C"

        # analyse file
        fps, dar, sar, max_frames, ac3_stream, error = self.analyse_mediafile(filename)
        if error:
            return None, "Konnte FPS nicht bestimmen: " + error
        # fix cutlist
        cutlist.fps = fps
        if not cutlist.cuts_frames:
            for start,dur in cutlist.cuts_seconds:
                cutlist.cuts_frames.append((int(0.5+(start*cutlist.fps)),int(0.5+(dur*cutlist.fps))))
        elif not cutlist.cuts_seconds:
            for start,dur in cutlist.cuts_frames:
                cutlist.cuts_seconds.append((float(start)/cutlist.fps),(float(dur)/cutlist.fps))

        # codec configuration string
        format, ac3_file, bframe_delay = self.get_format(filename)
        if format == Format.HQ:
            if encoder_engine == 'x264':
                codec, codec_core = self.complete_x264_opts(
                    self.config.get('smartmkvmerge', 'x264_hq_string').split(' '), filename)
            elif encoder_engine == 'ffmpeg':
                codec, codec_core = self.__ffmpeg_codec_options(
                    self.config.get('smartmkvmerge', 'ffmpeg_hq_x264_options').split(' '), filename)
        elif format == Format.HD:
            if encoder_engine == 'x264':
                codec, codec_core = self.complete_x264_opts(
                    self.config.get('smartmkvmerge', 'x264_hd_string').split(' '), filename)
            elif encoder_engine == 'ffmpeg':
                codec, codec_core = self.__ffmpeg_codec_options(
                    self.config.get('smartmkvmerge', 'ffmpeg_hd_x264_options').split(' '), filename)
##        elif format == Format.MP4:
##            if encoder_engine == 'x264':
##                codec, codec_core = self.complete_x264_opts(
##                    self.config.get('smartmkvmerge', 'x264_mp4_string').split(' '), filename)
##            elif encoder_engine == 'ffmpeg':
##                codec, codec_core = self.__ffmpeg_codec_options(
##                    self.config.get('smartmkvmerge', 'ffmpeg_mp4_x264_options').split(' '), filename)
##        elif format == Format.AVI:
##            encoder_engine = 'ffmpeg'
##            codec = self.config.get('smartmkvmerge', 'ffmpeg_avi_mpeg4_options').split(' ')
##            codec_core = 125
        else:
            return None, "Format nicht unterstützt (Nur MP4 H264, HQ H264 und HD H264 sind möglich)."

        if codec_core != 125:
            warning_msg ="Unbekannte Kodierung entdeckt. Diese Datei genau prüfen und notfalls mit intern-Virtualdub und Codec ffdshow schneiden."
            return None,  warning_msg
            
        # test workingdir
        if os.access(self.config.get('smartmkvmerge', 'workingdir').rstrip('/'),  os.W_OK):
            self.workingdir = os.path.abspath(self.config.get('smartmkvmerge', 'workingdir')).rstrip('/')
        else:
            return None, "Ungültiges Temp Verzeichnis. Schreiben nicht möglich."
          
        # threads  
        flag_singlethread = self.config.get('smartmkvmerge', 'single_threaded')
        
        if self.config.get('smartmkvmerge', 'single_threaded_automatic'):
            try:
                memory = self.meminfo()
                if self.available_cpu_count() > 1 and memory['MemFree'] > (os.stat(filename).st_size/1024):
                    flag_singlethread = True
                else:
                    flag_singlethread = False
            except Exception  as e:
                flag_singlethread = self.config.get('smartmkvmerge', 'single_threaded')
            
        # audio part 1 - cut audio 
        if ac3_file:
            audio_import_files.append(ac3_file)

        audio_timecodes = (',+'.join([self.get_timecode(start) + '-' + self.get_timecode(start+duration) for start, duration in cutlist.cuts_seconds]))        
        audio_timecodes = audio_timecodes.lstrip(',+')
        
        command = [mkvmerge, '--ui-language',  'en_US',  '-D',  '--split',  'parts:'+audio_timecodes,  '-o',  self.workingdir + '/audio_copy.mkv'] + audio_import_files
        try:
            blocking_process = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, universal_newlines=True,  env=my_env)
        except OSError as e:
            return None,  e.strerror + ": " + mkvmerge
        mkvmerge_list.append(blocking_process)
        if flag_singlethread:
            self.show_progress(blocking_process)

        # video part 1 - read keyframes
        keyframes, error = self.get_keyframes_from_file(filename)
        if keyframes == None:
            return None,  "Keyframes konnten nicht ausgelesen werden."
        # video part 2 - simulate smart rendering process
        for frame_start, frames_duration in cutlist.cuts_frames:
            result = self.__simulate_smart_mkvmerge(int(frame_start), int(frames_duration), keyframes)
            if result != None:
                videolist += result
            else:
                return None,  'Cutlist oder zu schneidende Datei passen nicht zusammen oder sind fehlerhaft.'
        # video part 3 - encode small parts - smart rendering part (1/2) 
        for encode, start,  duration,  video_part_filename in videolist:
            self.video_files.append('+'+ self.workingdir +'/' + video_part_filename)
            if encoder_engine == 'x264':
                command = [x264] + codec + ['--demuxer','ffms','--index', self.workingdir + '/x264.index','--seek',  str(start),'--frames',  str(duration),  '--output',  self.workingdir + '/' + video_part_filename,  filename ]
            elif encoder_engine == 'ffmpeg':
                command = [ffmpeg,  '-ss', str(self.get_timecode((start+bframe_delay)/fps)),  '-i',  filename, '-vframes',  str(duration), '-vf',  'setsar=' + str(sar), '-threads',  '0',  '-an',  '-sn',  '-dn', '-y',   self.workingdir + '/' + video_part_filename]
                command[5:5]=codec
            else:
                return None,  "Keine unterstützte Render-Engine zum Kodieren eingestellt"
            if encode:
                try:
                    non_blocking_process = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, universal_newlines=True)
                except OSError as e:
                    return None, e.strerror + ": " + 'Render Engine nicht vorhanden'
                process_list.append(non_blocking_process)
                if flag_singlethread:
                    self.show_progress(non_blocking_process)
            else:
                video_splitframes += ','+str(start)+'-'+str(duration)

        self.video_files[0]=self.video_files[0].lstrip('+')
        video_splitframes = video_splitframes.lstrip(',')
        
        # video part 4 - cut the big parts out the file (keyframe accurate) - smart rendering part (2/2)
        command = [mkvmerge,  '--ui-language',  'en_US','-A',  '--split',  'parts-frames:'+video_splitframes,  '-o',  self.workingdir + '/video_copy.mkv', filename ]
        try:
            non_blocking_process = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, universal_newlines=True, env=my_env)
        except OSError as e:
            return None, e.strerror + ": " + mkvmerge
        mkvmerge_list.append(non_blocking_process)
        if flag_singlethread:
            self.show_progress(non_blocking_process)

        # audio part 2 - encode audio to AAC
        if 'MP3 Spur kopieren' in self.config.get('smartmkvmerge', 'first_audio_stream') and 'AC3 Spur kopieren' in self.config.get('smartmkvmerge', 'second_audio_stream'):
            self.audio_files.append(self.workingdir + '/audio_copy.mkv')
        else:
                self.show_progress(blocking_process)
                blocking_process.wait()
                ffmpeginput_file = self.workingdir + '/audio_copy.mkv'
                ffmpegoutput_file = self.workingdir + '/audio_encode.mkv'
                
                audiofilter = []
                # convert first audio stream to aac
                if 'AAC' in self.config.get('smartmkvmerge', 'first_audio_stream') and 'AAC' in self.config.get('smartmkvmerge', 'second_audio_stream'):
                    aacaudiostreams = '-c:a'
                    if self.config.get('smartmkvmerge', 'normalize_audio'):
                        vol0,  error = self.get_norm_volume(ffmpeginput_file, '0')
                        vol1,  error = self.get_norm_volume(ffmpeginput_file, '1')
                        audiofilter = ['-af:0', 'volume=volume=' + vol0,  '-af:1', 'volume=volume=' + vol1]
                elif 'AAC' in self.config.get('smartmkvmerge', 'second_audio_stream') and 'MP3' in self.config.get('smartmkvmerge', 'first_audio_stream'):
                    aacaudiostreams = '-c:a:1'
                    if self.config.get('smartmkvmerge', 'normalize_audio'):
                        vol,  error = self.get_norm_volume(ffmpeginput_file, '1')
                        audiofilter = ['-af:1', 'volume=volume=' + vol]
                elif 'AAC' in self.config.get('smartmkvmerge', 'first_audio_stream'):
                    aacaudiostreams = '-c:a:0'
                    if self.config.get('smartmkvmerge', 'normalize_audio'):
                        vol,  error = self.get_norm_volume(ffmpeginput_file, '0')
                        audiofilter = ['-af:0', 'volume=volume=' + vol]
                else:
                    aacaudiostreams = '-c:a:2'
                    
                if 'nonfree' in ffmpeg:
                    # nonfree ffmpeg version with fdk support available
                    audiocodec = ['-c:a',  'copy',  aacaudiostreams,  'libfdk_aac',  '-flags',  '+qscale',  '-profile:a',  'aac_low',  '-global_quality',  '5' ,'-afterburner',  '1']
                else:
                    # only gpl version of ffmpeg available -> use standard aac codec
                    audiocodec = ['-c:a',  'copy',  aacaudiostreams,  'aac', '-strict', '-2','-profile:a',  'aac_low',  '-ab' ,'192k',  '-cutoff',  '18000']
                    
                if '2-Kanal' in self.config.get('smartmkvmerge', 'first_audio_stream'):
                    audiocodec.extend(['-ac:0',  '2'])

                if ac3_file == None:
                    # no ac3 stream found - all streams are muxed 
                    map = ['-map',  '0']
                else:
                    if 'AC3' in self.config.get('smartmkvmerge', 'first_audio_stream') :
                        map = ['-map',  '0:a:1']
                    else:
                        map = ['-map',  '0:a:0']
                    if not 'AC3 Spur entfernen' in self.config.get('smartmkvmerge', 'second_audio_stream') :
                        map.extend(['-map',  '0:a:1'])
                    
                args = [ffmpeg, "-loglevel", "info", "-y", "-drc_scale", "1.0", "-i", ffmpeginput_file, "-vn", "-vsync", "1", '-async',  '200000',  "-dts_delta_threshold", "100", '-threads',  '0',    ffmpegoutput_file]
                map.extend(audiocodec)
                map.extend(audiofilter)
                args[8:8] = map
                try:
                    non_blocking_process = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, universal_newlines=True)
                except OSError as e:
                    return None, e.strerror + ": " + ffmpeg
                process_list.append(non_blocking_process)
                self.audio_files.append(self.workingdir + '/audio_encode.mkv')
                if flag_singlethread:
                    self.show_progress(non_blocking_process)

        # wait until all threads are terminated
        for blocking_process in mkvmerge_list + process_list:
            self.show_progress(blocking_process)

        # check all processes
        for blocking_process in mkvmerge_list:
            returncode = blocking_process.wait()
            if returncode != 0 and returncode != 1:
                return None,  'beim Schneiden der Originaldatei...'
        for blocking_process in process_list:
            returncode = blocking_process.wait()
            if returncode != 0:
                return None,  'beim Kodieren ...'

        # clean up
        if os.path.isfile (self.workingdir + '/video_copy.mkv'):
            os.rename(self.workingdir + '/video_copy.mkv', self.workingdir + '/video_copy-001.mkv')
        if vars().has_key('ffmpeginput_file'):
            if os.path.isfile (ffmpeginput_file):
                os.remove(ffmpeginput_file)        
            
        # mux all together
        if self.config.get('smartmkvmerge', 'remux_to_mp4'):
            cut_video = self.workingdir + '/' + os.path.basename(os.path.splitext(self.generate_filename((filename),1))[0] + ".mkv")
        else:
            cut_video = os.path.splitext(self.generate_filename(filename,1))[0] + ".mkv"
        command = [mkvmerge,  '--engage', 'no_cue_duration', '--engage',  'no_cue_relative_position', '--ui-language',  'en_US',  '-o',  cut_video] + self.video_files + self.audio_files
        try:
            blocking_process = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, universal_newlines=True, env=my_env)
        except OSError:
            return None, "MKVMerge konnte nicht aufgerufen werden oder zu alt (6.5.0 benötigt)"
        self.show_progress(blocking_process)
        
        returncode = blocking_process.wait()
        if returncode != 0 and returncode != 1:
            return None,  'beim Schreiben des geschnittenen MKVs...'

        # remove all temporary files
        for n in self.video_files + self.audio_files:
            if os.path.isfile(n.lstrip('+')):
                os.remove(n.lstrip('+'))
                 
        return cut_video, warning_msg

    def __simulate_smart_mkvmerge(self, start,  duration,  keyframes):
        end = start + duration
        if start in keyframes:
            if (end) in keyframes:
                if end <= start:
                    return
                # copy keyframe to keyframe
                self.copy_nr +=1
                return [(False, start+1, end+1, 'video_copy-00'+str(self.copy_nr) + '.mkv')]
            else:
                # copy to keyframe before end
                try:
                    lt_kf_before_end = self.get_keyframe_in_front_of_frame(keyframes, end)
                except:
                    return None
                if lt_kf_before_end <= start:
                    self.encode_nr += 1
                    encode = [(True, start, duration, 'video_encode-00'+str(self.encode_nr)+'.mkv')]
                    return encode
                else:
                    self.copy_nr +=1
                    copy = [(False, start+1, lt_kf_before_end+1, 'video_copy-00'+str(self.copy_nr)+'.mkv')]
                    # encode to end of interval
                    self.encode_nr += 1
                    encode = [(True, lt_kf_before_end, end-lt_kf_before_end , 'video_encode-00'+str(self.encode_nr)+'.mkv')]
                    return copy + encode
        else:
            try:
                nt_kf_from_start = self.get_keyframe_after_frame(keyframes, start)
            except:
                return None
            duration_nt_kf = nt_kf_from_start-start
            if end <= nt_kf_from_start:
                self.encode_nr += 1
                encode = [(True, start, duration, 'video_encode-00'+str(self.encode_nr)+'.mkv')]
                return encode
            else:
                self.encode_nr += 1
                encode = [(True, start, duration_nt_kf, 'video_encode-00'+str(self.encode_nr)+'.mkv')]
                if  duration-duration_nt_kf > 0:
                    result = self.__simulate_smart_mkvmerge(nt_kf_from_start, duration-duration_nt_kf,  keyframes)
                    if result != None:
                        return encode + result
                    else:
                        return None
                else:
                    return encode

    def __ffmpeg_codec_options(self, ffmpeg_codec_options,  filename):
        codec = None
        codec_core = None
        ffmpeg_commandline = []
        bt709 = ['videoformat=pal:colorprim=bt709:transfer=bt709:colormatrix=bt709']
        bt470bg = ['videoformat=pal:colorprim=bt470bg:transfer=bt470bg:colormatrix=bt470bg']
        
        try:
            blocking_process = subprocess.Popen([self.config.get_program('mediainfo'), filename], stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True)
        except OSError as e:
            return None, "Fehler: %s Filename: %s Error: %s" % str(e.errno),  str(e.filename),  str(e.strerror)
        except ValueError as e:
            return None,  "Falscher Wert: %s" % str(e)
        
        while True:
            line = blocking_process.stdout.readline()

            if line != '':
                if 'x264 core' in line:
                    codec = 'libx264'
                    try:
                        codec_core = int(line.strip().split(' ')[30])
                    except ValueError as e:
                        continue
                    except IndexError as e:
                        continue
                elif 'Color primaries' in line and '709' in line:
                    ffmpeg_codec_options.extend(bt709)
                elif 'Color primaries' in line and '470' in line:
                    ffmpeg_codec_options.extend(bt470bg)
                elif 'Format profile' in line and '@L' in line:
                    try:
                        level = ['-level',  str(float(line.strip().split('L')[1]))] # test for float
                        profile = ['-profile:v',  line.strip().split('@L')[0].split(':')[1].lower().lstrip()]
                    except ValueError as e:
                        continue
                    except IndexError as e:
                        continue
                    ffmpeg_commandline.extend(profile)
                    ffmpeg_commandline.extend(level)
            else:
                break
                
        if codec == 'libx264':
            x264opts = (':'.join([option for option in ffmpeg_codec_options]))+':force_cfr'        
            x264opts = x264opts.lstrip(':')
            ffmpeg_commandline.extend(['-vcodec',  'libx264',  '-preset',  'medium',  '-x264opts',  x264opts])

        return ffmpeg_commandline,  codec_core

class ChangeDir:
    def __init__(self, newPath):
        self.newPath = newPath

    def __enter__(self):
        self.savedPath = os.getcwd()
        os.chdir(self.newPath)

    def __exit__(self, etype, value, traceback):
        os.chdir(self.savedPath)

class Init():
    def __init__(self,config):
        self.config = config
        self.msg = ''
    def main(self, setting=None):
        if not setting:
            self.setpaths()
            if self.msg:
                return self.msg
            self.setotr()
            if self.msg:
                return self.msg
            self.setencoder()
            if self.msg:
                return self.msg
            self.setthreads()
            if self.msg:
                return self.msg
            self.setoptions()
            return self.msg           
        elif setting == 'paths':
            self.setpaths()
            return self.msg
        elif setting == 'otr':
            self.setotr()
            return self.msg
        elif setting == 'threads':
            self.setthreads()
            return self.msg
        elif setting == 'encoder':
            self.setencoder()
            return self.msg
        elif setting == 'options':
            self.setoptions()
            return self.msg
        else:
            return "Unbekannter Parameter"
        
    def setpaths(self):
        dldir = self.config.get('general','folder_new_otrkeys')
        dldir_exists = os.path.exists(dldir)
        basepath = self.config.get('general','folder_uncut_avis').rsplit('/',1)[0]
        basepath_exists = os.path.exists(basepath)
        print 'Download-Verzeichnis für OTR-Keys und Cutlists'
        print 'aktueller Wert: ' + dldir
        if not dldir_exists:
            print 'Dieser Pfad ist ungültig und muss unbedingt gesetzt werden!'
        else:
            print 'Eine leere Eingabe bestätigt den aktuellen Pfad'
        print
        sel = raw_input('Download-Verzeichnis: ')
        if not sel and dldir_exists:
            sel = dldir
        else:
            while not sel or not os.path.exists(sel):
                print 'Dieser Pfad ist ungültig und muss unbedingt gesetzt werden!'
                print
                sel = raw_input('Download-Verzeichnis: ')
        self.config.setvalue('general','folder_new_otrkeys',sel.rstrip('/'))
        print
        print 'Basisverzeicnis für rpiotrtools mit verschiedenen Unterverzeichnissen'
        print 'aktueller Wert: ' + basepath
        if not basepath_exists:
            print 'Dieser Pfad ist ungültig und muss unbedingt gesetzt werden!'
        else:
            print 'Eine leere Eingabe bestätigt den aktuellen Pfad'
        print
        sel = raw_input('OTR-Verzeichnis: ')
        if not sel and basepath_exists:
            sel = basepath
        else:
            while not sel or not os.path.exists(sel):
                print 'Dieser Pfad ist ungültig und muss unbedingt gesetzt werden!'
                print
                sel = raw_input('OTR-Verzeichnis: ')
            basepath = sel.rstrip('/')
        decpath = os.path.join(basepath,'decoded')
        cutpath = os.path.join(basepath,'cut')
        trashpath = os.path.join(basepath,'trash')
        mkvpath = os.path.join(basepath,'converted')
        tmppath = os.path.join(basepath,'tmp')
        for path in [decpath,cutpath,trashpath,mkvpath,tmppath]:
            if not os.path.exists(path):
                try:
                    os.mkdir(path)
                except:
                    self.msg = "Verzeichnisse konnten nicht erzeugt werden"
        if not self.msg:
            self.config.setvalue('general','folder_uncut_avis',decpath)
            self.config.setvalue('general','folder_cut_avis',cutpath)
            self.config.setvalue('general','folder_trash_avis',trashpath)
            self.config.setvalue('general','folder_converted_avis',mkvpath)
            self.config.setvalue('smartmkvmerge','workingdir',tmppath)

    def setotr(self):
        email = self.config.get('general','email')
        pw = self.config.get('general','password')
        print 'Der Dekoder benötigt die bei OTR hinterlegte Emailadresse.'
        print 'Ist sie nicht gespeichert, wird sie beim Dekodieren abgefragt.'
        if email:
            print 'Aktuelle Einstellung: ' + email
            print "Eine leere Eingabe löscht die aktuelle Einstellung"
        else:
            print "Aktuell nicht gespeichert"
        print
        sel = raw_input('OTR-Verzeichnis: ')
        if email and not sel:
            print "Aktuelle Email-Adresse wird gelöscht, muss manuell eingegeben werden"
        elif sel:
            print "Email-Adresse wird gespeichert"
        else:
            print "Email-Adresse nicht gespeichert, muss manuell eingegeben werden"
        self.config.setvalue('general','email',sel)
        print
        print 'Der Dekoder benötigt das bei OTR hinterlegte Passwort.'
        print 'Ist es nicht gespeichert, wird es beim Dekodieren abgefragt.'
        print 'Das Speichern des Passwortes ist ein hohes Sicherheitsrisiko!'
        if pw:
            print "Es ist aktuell ein Passwort gesetzt"
            print "Eine leere Eingabe löscht die aktuelle Einstellung"
        else:
            print "Aktuell nicht gespeichert"
        print
        sel = raw_input('OTR-Passwort: ')
        if pw and not sel:
            print "Aktuelles Passwort wird gelöscht, muss manuell eingegeben werden"
        elif sel:
            print "Passwort wird gespeichert"
        else:
            print "Passwort nicht gesetzt, muss manuell eingegeben werden"
        self.config.setvalue('general','password',sel)

    def setthreads(self):
        single_threaded = self.config.get('smartmkvmerge','single_threaded')
        print "Smartmkvmerge kann single- oder multithreaded arbeiten."
        print "Multithreaded geht etwas schneller, kann aber sehr viel "
        print "Speicherplatz benötigen und den RPi zum Swappen veranlassen."
        if single_threaded:
            print "Aktuelle Einstellung: Single-Threaded"
        else:
            print "Aktuelle Einstellung: Multi-Threaded"
        print
        sel = ''
        while sel not in ['s','m']:
            sel = raw_input("Threading 's' oder 'm': ")
        if sel == 's':
            self.config.setvalue('smartmkvmerge','single_threaded',True)
            self.config.setvalue('smartmkvmerge','single_threaded_automatic',False)
        else:
            self.config.setvalue('smartmkvmerge','single_threaded',False)
            self.config.setvalue('smartmkvmerge','single_threaded_automatic',True)

    def setencoder(self):
        encoder_engine = self.config.get('smartmkvmerge','encoder_engine')
        print "Es kann x264 oder ffmpeg für das Smart Rendering verwendet werden."
        print "ffmpeg läuft etwas schneller, benötigt aber mehr Arbeitsspeicher."
        print "Es sollte deshalb nur im Single-Threading-Modus verwendet werden."
        print "Empfohlen wird die Verwendung von x264."
        print "Aktuelle Einstellung: " + encoder_engine
        print
        sel = ''
        while sel not in ['x','f']:
            sel = raw_input("Encoder 'x' oder 'f': ")
        if sel == 'x':
            self.config.setvalue('smartmkvmerge','encoder_engine','x264')
        else:
            self.config.setvalue('smartmkvmerge','encoder_engine','ffmpeg')

    def setoptions(self):
        delete = self.config.get('general','delete_otrkeys')
        rename = self.config.get('general','rename_cut')
        play = self.config.get('general','play_cut_videos')
        print "OTR-Keys nach dem Dekdodieren ohne Rückfrage löschen?"
        if delete:
            print "Aktuelle Einstellung: Ja"
        else:
            print "Aktuelle Einstellung: Nein"
        print
        sel = ''
        while sel not in ['j','n']:
            sel = raw_input("Automatisch löschen? 'j' oder 'n': ")
        if sel == 'j':
            self.config.setvalue('general','delete_otrkeys',True)
        else:
            self.config.setvalue('general','delete_otrkeys',False)
        print

        print "Dateinamen der geschnittenen Datei (hinter Laufzeit) kürzen?"
        if rename:
            print "Aktuelle Einstellung: Ja"
        else:
            print "Aktuelle Einstellung: Nein"
        print
        sel = ''
        while sel not in ['j','n']:
            sel = raw_input("Dateinamen kürzen? 'j' oder 'n': ")
        if sel == 'j':
            self.config.setvalue('general','rename_cut',True)
        else:
            self.config.setvalue('general','rename_cut',False)
        print

        print "Geschnittene Videos nach dem Schneiden abspielen?"
        print "Benötigt omxplayerGUI oder VLC."
        if play:
            print "Aktuelle Einstellung: Ja"
        else:
            print "Aktuelle Einstellung: Nein"
        print
        sel = ''
        while sel not in ['j','n']:
            sel = raw_input("Geschnittene Videos abspielen? 'j' oder 'n': ")
        if sel == 'j':
            self.config.setvalue('general','play_cut_videos',True)
        else:
            self.config.setvalue('general','play_cut_videos',False)
     
def copycutlists(config):
    dldir = config.get('general','folder_new_otrkeys')
    uncutdir = config.get('general','folder_uncut_avis')
    if os.path.exists(dldir) and os.path.isdir(dldir) and os.path.exists(uncutdir) and os.path.isdir(uncutdir):
        cutlists = []
        dirlist = os.listdir(dldir)
        for f in dirlist:
            if f.endswith('.cutlist'):
                cutlists.append(f)
        if cutlists:
            for cl in cutlists:
                shutil.move(os.path.join(dldir,cl),os.path.join(uncutdir,cl))
            print str(len(cutlists)) + ' Cutlists kopiert'

def decode(config):
    delete_otrkeys = config.get('general','delete_otrkeys')
    mailadr = config.get('general','email')
    pw = config.get('general','password')
    decoder = config.get_program('decoder')
    if not mailadr:
        mailadr = raw_input("OTR Emailadresse : ")
    if not pw:
        pw = raw_input("OTR Passwort : ")
    dldir = config.get('general','folder_new_otrkeys')
    uncutdir = config.get('general','folder_uncut_avis')
    if mailadr and pw and os.path.exists(dldir) and os.path.isdir(dldir) and os.path.exists(uncutdir) and os.path.isdir(uncutdir):
        exe = [decoder,'-d','-e',mailadr,'-p',pw,'-D', uncutdir]
        files = os.listdir(dldir)
        keys = []
        decoded = []
        for f in files:
            if f.endswith('.otrkey'):
                movie = f.replace('.otrkey','')
                if not os.path.exists(os.path.join(uncutdir,movie)):
                    keys.append(os.path.join(dldir,f))
        if keys:
            keys.sort()
            for k in keys:
                print k.split('/')[-1]
                cmd = exe + [k]
                proc = subprocess.Popen(cmd)
                res = proc.wait()
                if res == 0:
                    if delete_otrkeys:
                        os.remove(k)
                    else:
                        decoded.append(k)
                else:
                    print "Fehler beim Dekodieren von "+k
            if decoded and not delete_otrkeys:        
                sel = raw_input("OTR-Keys löschen? (y,j) : ")
                if sel in ['j','J','y','Y']:
                    for k in decoded:
                        os.remove(k)
        else:
            print 'Keine OTR-Keys gefunden'
    else:
        print "Fehler: Konnte nichts dekodieren."

def clean(config):
    uncutdir = config.get('general','folder_uncut_avis')
    files = os.listdir(uncutdir)
    for f in files:
        if f.endswith('.kf.txt'):
            os.remove(os.path.join(uncutdir,f))

def cutmovie(config,moviefile,cutlistfile):
    global cutvideos
    uncutdir = config.get('general','folder_uncut_avis')
    trashdir = config.get('general','folder_trash_avis')
    moviepath = os.path.join(uncutdir,moviefile)
    cutlistpath = os.path.join(uncutdir,cutlistfile)
    cutlist = Cutlist(cutlistpath)
    cutlist.read_from_file()
    res = cutlist.read_cuts()
    cutit = True
    if res == None:
        if cutlist.filename_original != moviefile:
            start = moviefile.split("TVOON_DE")[0]+"TVOON_DE"
            if cutlist.filename_original.startswith(start):
                cutlist.cuts_frames = []
            else:
                cutit = False
                res = "Fehler: Cutlist passt nicht zur Filmdatei"
        if cutit:
            print "Schneide: " + moviepath.split('/')[-1]
            cutter = CutSmartMkvmerge(config)
            name,err = cutter.cut_file_by_cutlist(moviepath,cutlist )
            if name:
                if config.get("general","rename_cut") and "_TVOON_DE" in name:
                    newname = name.split('_TVOON_DE')[0] + ".mkv"
                    try:
                        os.rename(name,newname)
                        name = newname
                    except:
                        pass
                cutvideos.append(name)
                res = "Erfolgreich geschnittene Datei: " +name.split('/')[-1]
                shutil.move(moviepath,os.path.join(trashdir,moviefile))
                shutil.move(cutlistpath,os.path.join(trashdir,cutlistfile))
                if moviefile.endswith('.HQ.avi'):
                    ac3file = moviefile.replace('.HQ.avi','.HD.ac3')
                elif moviefile.endswith('.HD.avi'):
                    ac3file = moviefile.replace('.avi','.ac3')
                else:
                    ac3file = ''
                if ac3file and os.path.exists(os.path.join(uncutdir,ac3file)):
                    shutil.move(os.path.join(uncutdir,ac3file),os.path.join(trashdir,ac3file))              
            else:
                res = 'Fehler: ' + err
    print res

def cut(config,moviefile=None,cutlistfile=None):
    uncutdir = config.get('general','folder_uncut_avis')
    if moviefile and cutlistfile:
        if os.path.exists(os.path.join(uncutdir,moviefile)) and os.path.exists(os.path.join(uncutdir,cutlistfile)):
            cutmovie(config,moviefile,cutlistfile)
        else:
            print "Datei nicht gefunden"
    else:
        todo = []
        files = os.listdir(uncutdir)
        files.sort()
        if moviefile:
            if os.path.exists(os.path.join(uncutdir,moviefile)):
                if moviefile+'.cutlist' in files:
                    todo.append((moviefile,moviefile+'.cutlist'))
                else: #cutlist mit anderen Namen, z.b. von OTR
                    start = moviefile.split("TVOON_DE")[0]+"TVOON_DE"
                    for c in files:
                        if c.endswith('.cutlist') and c.startswith(start):
                            todo.append((f,c))
        else:
            for f in files:
                if f.endswith('.avi'):
                    if f+'.cutlist' in files:
                        todo.append((f,f+'.cutlist'))
                    else:
                        start = f.split("TVOON_DE")[0]+"TVOON_DE"
                        for c in files:
                            if c.endswith('.cutlist') and c.startswith(start):
                                todo.append((f,c))
        if todo:
            for moviefile,cutlistfile in todo:
                cutmovie(config,moviefile,cutlistfile)
        else:
            print "Keine Filmdateien bzw. passende Cutlists gefunden"
    clean(config)

def convert(config,moviefile=None):
    todo = []
    done = []
    uncutdir = config.get('general','folder_uncut_avis')
    mkvdir = config.get('general','folder_converted_avis')
    trashdir = config.get('general','folder_trash_avis')
    files = os.listdir(uncutdir)
    files.sort()
    dfiles = os.listdir(uncutdir)
    if moviefile:
        if os.path.exists(os.path.join(uncutdir,moviefile)) and moviefile.endswith('.avi') and moviefile.replace('.avi','.mkv') not in dfiles:
            todo.append(moviefile)
    else:
        files = os.listdir(uncutdir)
        files.sort()
        dfiles = os.listdir(mkvdir)
        for f in files:
            if f.endswith('.avi') and f.replace('.avi','.mkv') not in dfiles:
                todo.append(f)
    if todo:
        for f in todo:
            src = os.path.join(uncutdir,f)
            dst = os.path.join(mkvdir,f.replace('.avi','.mkv'))
            ret_value = -1
            try:
                proc = subprocess.Popen(['mkvmerge','-q','-o',dst,src])
                print "Konvertiere: " + f
                ret_value = proc.wait()
            except:
                pass
            if ret_value == 0:
                done.append(f)
            else:
                print "Fehler beim Konvertieren!"
        if done:
            sel = raw_input("Sollen die konvertierten Dateien in den Trash-Ordner verschoben werden? (j,n):")
            if sel in ['j','J']:
                for d in done:
                    src = os.path.join(uncutdir,d)
                    dst = os.path.join(trashdir,d)
                    try:
                        shutil.move(src,dst)
                    except:
                        pass
    else:
         print "Keine Datei(en) zum Konvertieren gefunden."
                
def checkpaths(config):
    res = True
    dpaths = ["folder_cut_avis","folder_new_otrkeys","folder_trash_avis","folder_uncut_avis","folder_converted_avis"]
    for d in dpaths:
        path = config.get("general",d)
        if not (os.path.exists(path) and os.path.isdir(path)):
            res = False
            break
    return res

def set_options(config,options):
    for opt in options:
        if opt == '-m':
            config.setvalue('smartmkvmerge','single_threaded',False)
            config.setvalue('smartmkvmerge','single_threaded_automatic',True)
        elif opt == '-s':
            config.setvalue('smartmkvmerge','single_threaded',False)
            config.setvalue('smartmkvmerge','single_threaded_automatic',True)
        elif opt == '-l':
            config.setvalue('general','delete_otrkeys',True)
        elif opt == '-a':
            config.setvalue('general','play_cut_videos',True)
        elif opt == '-r':
            config.setvalue('general','rename_cut',True)

def emptytrash(config):
    trashdir = config.get('general','folder_trash_avis')
    if os.path.exists(trashdir) and os.path.isdir(trashdir):
        files = os.listdir(trashdir)
        for f in files:
            os.remove(os.path.join(trashdir,f))
        print str(len(files)) + " Dateien gelöscht."
    else:
        print "Verzeichnis nicht gefunden"
   
def init(config,setting=None):
    i = Init(config)
    msg = i.main(setting)
    if msg:
        print msg
    else:
        config.save(homedir+'/.rpiotrtool.conf')
    sys.exit(0)

def usage():
    print "rpiotrtool Vers. 0.5 - Copyright 2018 by Günter Kreidl"
    print "Aufruf: rpiotrtool [Optionen] [Kommando] [Wert1] [Wert2]"
    print
    print "Verfügbare Kommandos:"
    print "decode"
    print "Dekodiert alle neuen OTR-Keys"
    print 
    print "cut [Datei] [Cutlist]"
    print "Ohne weitere Argumente: Schneidet alle Dateien für die eine Cutlist vorhanden ist."
    print "Man kann auch eine bestimmte Datei und evtl. Cutlist zum Schneiden angeben."
    print
    print "convert [Datei]"
    print "Konvertiert eine oder alle ungeschnittenen Dateien ins MKV-Format"
    print
    print "init [Sektion]"
    print "Ohne weiteres Agument werden all möglichen Einstellungen abgefragt und gesetzt."
    print "Verfügbare Sektionen:"
    print "'paths': Verzeichniseinstellungen"
    print "'otr': OTR-Einstellungen für Dekoder"
    print "'encoder': Wähle Encoder für das Smart Rendering (x264 oder ffmpeg)"
    print "'threads': Single- oder Multithreading beim Schneiden"
    print "'options': diverse Einstellungen"
    print
    print "clean"
    print "Löscht alle Dateien im Trash-Verzeichnis"
    print
    print "get"
    print "Holt alle neuen Cutlists aus dem Download-Verzeichnis"
    print "Dies geschieht auch automatisch bei jedem 'cut' oder 'decode' Kommando"
    print
    print "help"
    print "Zeigt diese Hilfe an. Funktioniert auch mit '-h' und '--help'"
    print
    print "Wird das Programm ohne jedes Argument gstartet, werden folgende Kommandos"
    print "nacheinander aufgerufen: get, decode, cut"
    print
    print "Optionen:"
    print "'-l': Lösche dekodierte OTR-Keys ohne Rückfrage"
    print "'-s': Single-Threaded schneiden"
    print "'-m': Multi-Threaded schneiden"
    print "'-r': Dateinamen der geschnittenen Datei kürzen"
    print "'-a': Videos nach dem Schneiden anzeigen"

    sys.exit(0)

def playvideos(config):
    global cutvideos
    if not config.get("general","play_cut_videos"):
        return
    if not cutvideos:
        return
    prefplayer = "" 
    for player in config.get("general","videoplayers"):
        if os.path.exists(player):
            prefplayer = player
            break
    if prefplayer:
        for video in cutvideos:
            subprocess.call([prefplayer,video])
    else:
        print "Keinen geeigneten Videoplayer gefunden"

cutvideos = []
version = "0.5"
allowed_commands = ['decode','cut','convert','init', 'clean', 'get', 'help','-h','--help',]
allowed_options = ['-m','-s','-a','-l','-r']
init_options = ['paths','otr','threads','options','encoder']
homedir = os.path.expanduser('~')
config = Config()
if os.path.exists(homedir+'/.rpiotrtool.conf'):
    config.load(homedir+'/.rpiotrtool.conf')
else:
    config.save(homedir+'/.rpiotrtool.conf')

if not checkpaths(config):
    init(config,"paths")
args = []
options = []

for arg in sys.argv[1:]:
    if arg in allowed_options:
        options.append(arg)
    else:
        args.append(arg)

if not args:
    set_options(config,options)
    copycutlists(config)
    decode(config)
    cut(config)
    playvideos(config)
else:
    cmd = args[0]
    if not cmd in allowed_commands:
        usage()
    if cmd in ['help','-h','--help',]:
        usage()
    elif cmd == "decode":
        set_options(config,options)
        copycutlists(config)
        decode(config)
    elif cmd == "get":
        copycutlists(config)
    elif cmd == "init":
        if len(args) == 1:
            init(config)
        elif len(args) > 1:
            setting = args[1]
            if setting in init_options:
                init(config,setting)
            else:
                usage
    elif cmd == "clean":
        emptytrash(config)
    elif cmd == "cut":
        set_options(config,options)
        copycutlists(config)
        if len(args) == 1:
            cut(config)
            playvideos(config)
        elif len(args) == 2:
            cut(config,args[1])
            playvideos(config)
        elif len(args) > 2:
            cut(config,args[1],args[2])
            playvideos(config)
    elif cmd == "convert":
        set_options(config,options)
        if len(args) == 1:
            convert(config)
        elif len(args) > 1:
            convert(config,args[1])

sys.exit(0)
